//
//  GSShape.h
//  GlyphsCore
//
//  Created by Georg Seifert on 06.04.19.
//  Copyright © 2019 schriftgestaltung.de. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <GlyphsCore/GSBase.h>
#import <GlyphsCore/GSContainerProtocol.h>
#import <GlyphsCore/GSSelectableElementProtocol.h>

typedef enum : unsigned short {
	GSShapeTypePath = 1 << 1,
	GSShapeTypeComponent = 1 << 2,
	GSShapeTypeGroup = 1 << 3,
} GSShapeType;

typedef NS_ENUM(NSUInteger, GSLineCapStyle) {
	GSLineCapStyleButt = 0,
	GSLineCapStyleRound = 1,
	GSLineCapStyleRoundInset = 2,
	GSLineCapStyleSquare = 3,
	GSLineCapStyleAlignToAxis = 4,
};

NS_ASSUME_NONNULL_BEGIN

extern NSString *const GSPathFillColorKey;
extern NSString *const GSPathFillKey;
extern NSString *const GSPathFillPatternFileKey;
extern NSString *const GSPathFillPatternBlendModeKey;
extern NSString *const GSPathFillPatternOffsetKey;
extern NSString *const GSPathFillPatternScaleKey;
extern NSString *const GSPathFillPatternAngleKey;
extern NSString *const GSPathGradientKey;
extern NSString *const GSPathStrokeColorKey;
extern NSString *const GSPathStrokeWidthKey;
extern NSString *const GSPathStrokeHeightKey;
extern NSString *const GSPathStrokePositionKey;
extern NSString *const GSPathStrokeLineCapStartKey;
extern NSString *const GSPathStrokeLineCapEndKey;
extern NSString *const GSPathStrokeLineJoinKey;
extern NSString *const GSPathShadowKey;
extern NSString *const GSPathInnerShadowKey;
extern NSString *const GSPathNoiseKey;
extern NSString *const GSShapeMaskKey;
extern NSString *const GSComponentReversePathsKey;

extern NSString *const GSAttributesKey;
extern NSString *const GSPathKey;

extern NSString *const GSPathMergeOpenPaths;

extern NSString *const GSShapeIdentifier;

@protocol GSPenProtocol;

@interface GSShape : NSObject < /*NSCoding, */ NSCopying, GSSelectableElementProtocol> {
#ifdef DEBUG
	__unsafe_unretained GSContainer* _parent;
#else
	__weak GSContainer* _parent;
#endif
#pragma GSAutoCodeStart ivars

	NSMutableDictionary *_attributes;

#pragma GSAutoCodeEnd ivars

@public
	BOOL _locked;
	NSPoint _position;
	NSPoint _positionPrecise;
}

/** initialises a shape with a dictionary loaded from a pList.
 
 @param dict A dictionary
 */
- (instancetype)initWithDict:(NSDictionary *)dict format:(GSFormatVersion)formatVersion;

- (BOOL)postRead:(NSError **)error format:(GSFormatVersion)formatVersion;

- (BOOL)isEqualToShape:(GSShape *)other;

/// The position of the element.
@property (assign, nonatomic) NSPoint position;

/** The unrounded position of the element.
 
 This is used for consecutiv operations to avoid rounding errors
 */
@property (nonatomic, readonly) NSPoint positionPrecise;

/** Sets the position without notifying the path and no undo.
 
 This is mostly used to build a new element list that is later added to the layer. Otherwise you need to make sure to anounce the changes. e.g. by calling `[GSLayer elementDidChange:]`.
 
 @param position The new position
 */

- (void)setPositionFast:(NSPoint)position;

- (void)setPositionFast:(NSPoint)position round:(CGFloat)grid;

- (void)transformFast:(NSAffineTransform *)transform round:(CGFloat)gridLength;

- (void)transform:(NSAffineTransform *)transform;

- (void)transform:(NSAffineTransform *)transform selection:(nullable NSOrderedSet *)selection;

- (void)getPositionsFromShape:(GSShape *)otherPath;

/**
 Moves the path by the offset.

 @param offset A NSPoint
*/
- (void)moveWithPoint:(NSPoint)offset;

@property (nonatomic) BOOL locked;

- (void)makeUnlocked;

- (void)makeLocked;

- (void)elementDidChange:(id)element;

#ifndef GLYPHS_VIEWER
/// Returns the undoManager of the containing glyph.
- (nullable NSUndoManager *)undoManager;
#endif

@property (readonly) GSShapeType shapeType;

- (void)scaleAttributes:(CGFloat)scale;

- (BOOL)getStrokeSettingsWidth:(CGFloat *_Nullable)width height:(CGFloat *)height pos:(CGFloat *_Nullable)pos capStart:(GSLineCapStyle *_Nullable)capStart capEnd:(GSLineCapStyle *_Nullable)capEnd join:(NSLineJoinStyle *_Nullable)join fill:(BOOL *_Nullable)fill mask:(BOOL *_Nullable)mask;
#ifndef GLYPHS_VIEWER
/**
 writes that shape to file
 
 @param file Nullable for testing purposes.
 @param formatVersion The file format version
 @param error error
 @return return value description
 */
- (BOOL)saveToFile:(FILE *)file format:(GSFormatVersion)formatVersion error:(NSError **)error;
#endif

/** Returns the content of the object to store it in pList.
 
 This is used to store the data in the .glyphs file.
 @param format the version of the dict
 */
- (nullable NSDictionary *)propertyListValueFormat:(GSFormatVersion)format;

- (NSRect)bounds;

- (void)drawInPen:(NSObject<GSPenProtocol> *_Nonnull)pen;

- (void)drawInPen:(NSObject<GSPenProtocol> *_Nonnull)pen openPen:(NSObject<GSPenProtocol> *__nullable)openPen secondaryPen:(NSObject<GSPenProtocol> *__nullable)secondPen extraHandles:(NSMutableArray *__nullable)extraHandles;

- (void)drawSimpleInPen:(NSObject<GSPenProtocol> *_Nonnull)pen;

#pragma GSAutoCodeStart methods

#pragma mark Attributes

/// @name Attributes
/** Place to store data.
 
 Here it is possible to store something. Please use a unique key.
 The objects should be able to serialize to a plist. If they are not, they should return all infos to store in description (as String).
 */
@property (strong, nonatomic, nullable) NSDictionary *attributes;

- (void)setAttributes:(NSDictionary * _Nullable)attributes format:(GSFormatVersion)format;

/**
 convenience accessor to get to the content of the attributes dict
 
 @param key the key
 @return the data stored with key
 */
- (NSObject * _Nullable)attributeForKey:(NSString *)key;


/** Adds something to the fonts attributes.
 
 This also triggers undo/document dirty state.
 
 @param value The object should be able to serialize to a plist. If they are not, they should return all infos to store in description (as String).
 @param key   Please use an uniqe key that gives some indecation who put it there (e.g. prefix it with your name or project).
 */
- (void)setAttribute:(nullable id)value forKey:(nonnull NSString *)key;


/** Removed the object with this key.
 
 @param key The key to remove
 */
- (void)removeAttributeForKey:(NSString *)key;
#pragma GSAutoCodeEnd methods

@end

NS_ASSUME_NONNULL_END
